/*
 *	The Lestrade Search Library -- Index File Format
 *
 *	(c) 2004--2005 Martin Mares <mj@ucw.cz>
 *	(c) 2005 Tomas Valla <tom@ucw.cz>
 */

#ifndef _LESTRADE_INDEX_H
#define _LESTRADE_INDEX_H

/*
 *  The index files have the following format:
 *
 *	header			struct index_header
 *	custom data		private data passed through from/to custom routines
 *	card attributes		struct card_attr [num_cards]	<-- defined in custom.h
 *	reference chains	sequence {		for all words/complexes
 *					sequence {		for all objects containing the word
 *						u32  header
 *							where the lowest 4bits is length1 and the rest (header>>4) is 28bit oid
 *								length1:  1..15 = direct length (99%), 0 = use length2 (1%)
 *						if (!length1) {
 *							utf8 length2		of ref[]
 *						}
 *						reference ref[count]
 *					}
 *					u32  0		end marker
 *				}
 *
 *	lexicon pointers	sequence [MAX_WORD_CHARS+1] {
 *					u32 word_start[]	<-- words of this size sorted lexicographically,
 *								    first without accents, then with them
 *				}
 *	lexicon			sequence [] {
 *					byte length;		<-- word length in bytes
 *					byte word[length]	<-- UTF-8 representation of the word
 *					byte pos[BYTES_PER_O]	<-- start of the reference chain (relative to refs_start)
 *				}
 */

/*
 * The reference format is following:
 *
 * union {
 * 	(a) 0tpp pppp                   2 most frequent word-types, 6-bit delta
 * 	(b) 10tt tppp pppp pppp         word-type, 11-bit delta, for word-types <= 7
 * 	(c) 11tt ttpp pppp pppp         word-type, 18-bit delta
 * 	    u8 pos
 * }
 */


#ifdef CONFIG_LARGE_DB
typedef u64 index_pos_t;
#define BYTES_PER_O 5
#define GET_O(p) GET_U40(p)
#define PUT_O(p,x) PUT_U40(p,x)
#define bgeto(f) bget5(f)
#define bputo(f,x) bput5(f,x)
#else
typedef u32 index_pos_t;
#define BYTES_PER_O 4
#define GET_O(p) GET_U32(p)
#define PUT_O(p,x) PUT_U32(p,x)
#define bgeto(f) bgetl(f)
#define bputo(f,x) bputl(f,x)
#endif

struct index_header {
  u32 magic;				/* INDEX_MAGIC_xxx */
  u32 version;				/* INDEX_VERSION */
  u32 num_cards;			/* Number of cards indexed */
  u32 timestamp;			/* Time when the index was built */
  index_pos_t custom_start;		/* Positions of parts of the index */
  index_pos_t card_attr_start;
  index_pos_t refs_start;
  index_pos_t lexicon_ptrs_start;
  index_pos_t lexicon_start;
  u32 lexicon_zone[MAX_WORD_CHARS+2];	/* Index of the first lexicon pointer for each word length */
};

#define INDEX_MAGIC 0x9f175fa4		/* Finished index file */
#define INDEX_TEMP_MAGIC 0x4af571f9	/* Index under construction */

#define INDEX_FORMAT_VERSION 0x01	/* Current index format version */
#ifdef CONFIG_LARGE_DB
#define INDEX_FORMAT_LARGE 0x01
#else
#define INDEX_FORMAT_LARGE 0x00
#endif
#define INDEX_VERSION ((INDEX_FORMAT_VERSION << 24) | (INDEX_FORMAT_LARGE << 16) | (CUSTOM_INDEX_TYPE << 8) | CUSTOM_INDEX_VERSION)

#endif
