/*
 *	The Lestrade Search Library -- Configuration & Initialization
 *
 *	(c) 2004--2005 Martin Mares <mj@ucw.cz>
 *	(c) 2004--2005 Tomas Valla <tom@ucw.cz>
 */

#include "ucw/lib.h"
#include "ucw/conf.h"
#include "ucw/getopt.h"
#include "ucw/simple-lists.h"
#include "lestrade/internal.h"
#include "lestrade/alphabet.h"

uns lestrade_initialized;
uns lex_max_gap = 3;
uns lex_max_ctrl_len = 16;
uns lex_max_hex_len = 10;
uns lex_max_len = 32;
uns lex_min_len_ign = 2;
uns lex_min_len = 2;
uns max_words = HARD_MAX_WORDS;
uns max_phrases = HARD_MAX_PHRASES;
uns max_nears = HARD_MAX_NEARS;
uns max_bools = HARD_MAX_BOOLS;
uns max_occurs = HARD_MAX_OCCURS;
uns max_phrase_len = HARD_MAX_PHRASE_LEN;
uns accent_penalty = 20;
uns default_weight = 10000;
uns word_weight_scale = 10;
uns near_bonus_word = 700;
uns near_penalty_gap = 500;
uns near_bonus_connect = 200;
uns morph_lemma_penalty = 20;
uns morph_basic_penalty = 25;
uns morph_other_penalty = 30;
uns morph_preacc_penalty = 20;
uns wild_min_prefix = 3;
uns wild_max_variants = 256;
uns wild_silent_fail;
uns wild_max_cached = 256;

clist lex_exceptions;

static clist lex_words_ignored;
static clist lex_words_normal;
static clist lex_words_garbage;

static void add_exceptions(struct clist *list, int class)
{
	CLIST_FOR_EACH(simp_node *, n, *list) {
		struct exception *e = cf_malloc(sizeof(struct exception));
		e->w = n->s;
		e->class = class;
		clist_add_tail(&lex_exceptions, &e->n);
	}
}

static char *lex_commit(void *p UNUSED)
{
	CF_JOURNAL_VAR(lex_exceptions);
	clist_init(&lex_exceptions);
	add_exceptions(&lex_words_ignored, WC_IGNORED);
	add_exceptions(&lex_words_normal, WC_NORMAL);
	add_exceptions(&lex_words_garbage, WC_GARBAGE);
	return NULL;
}

static struct cf_section lestrade_search_config = {
	CF_ITEMS {
		CF_UNS("MaxWords",		&max_words),
		CF_UNS("MaxPhrases",		&max_phrases),
		CF_UNS("MaxNears",		&max_nears),
		CF_UNS("MaxBools",		&max_bools),
		CF_UNS("MaxOccurs",		&max_occurs),
		CF_UNS("MaxPhraseLen",		&max_phrase_len),
		CF_UNS("AccentPenalty",		&accent_penalty),
		CF_UNS("DefaultWeight",		&default_weight),
		CF_UNS("WordWeightScale",	&word_weight_scale),
		CF_UNS("NearBonusWord",		&near_bonus_word),
		CF_UNS("NearPenaltyGap",	&near_penalty_gap),
		CF_UNS("NearBonusConnect",	&near_bonus_connect),
		CF_UNS("MorphLemmaPenalty",	&morph_lemma_penalty),
		CF_UNS("MorphBasicPenalty",	&morph_basic_penalty),
		CF_UNS("MorphOtherPenalty",	&morph_other_penalty),
		CF_UNS("PreaccentPenalty",	&morph_preacc_penalty),
		CF_UNS("WildMinPrefix",		&wild_min_prefix),
		CF_UNS("WildMaxMatches",	&wild_max_variants),
		CF_UNS("WildSilentFail",	&wild_silent_fail),
		CF_UNS("WildMaxCached",		&wild_max_cached),
		CF_END
	}
};

static struct cf_section lestrade_lexicon_config = {
	CF_COMMIT(lex_commit),
	CF_ITEMS {
		CF_UNS("MinWordLenIgnore",	&lex_min_len_ign),
		CF_UNS("MinWordLen",		&lex_min_len),
		CF_UNS("MaxWordLen",		&lex_max_len),
		CF_UNS("MaxHexWordLen",		&lex_max_hex_len),
		CF_UNS("MaxCtrlWordLen",	&lex_max_ctrl_len),
		CF_UNS("MaxGap",		&lex_max_gap),
		CF_LIST("WordIgnored",		&lex_words_ignored,	&cf_string_list_config),
		CF_LIST("WordNormal",		&lex_words_normal,	&cf_string_list_config),
		CF_LIST("WordGarbage",		&lex_words_garbage,	&cf_string_list_config),
		CF_END
	}
};

static void CONSTRUCTOR
lestrade_preinit(void)
{
	clist_init(&lex_exceptions);
	cf_declare_section("Search", &lestrade_search_config, 0);
	cf_declare_section("Lexicon", &lestrade_lexicon_config, 0);
}

void lestrade_init(byte *config_file)
{
	if (config_file)
		cf_load(config_file);
	else
		cf_set("");
	ASSERT(max_words <= HARD_MAX_WORDS);
	ASSERT(max_phrases <= HARD_MAX_PHRASES);
	ASSERT(max_nears <= HARD_MAX_NEARS);
	ASSERT(max_bools <= HARD_MAX_BOOLS);
	ASSERT(max_occurs <= HARD_MAX_OCCURS);
	ASSERT(max_phrase_len <= HARD_MAX_PHRASE_LEN);
	alphabet_init();
	words_global_init();
	full_global_init();
	details_global_init();
	lestrade_initialized = 1;
}

void lestrade_cleanup(void)
{
	lestrade_initialized = 0;
}
