/*
 *	The Lestrade Search Library -- Internal Stuff
 *
 *	(c) 2004--2005 Martin Mares <mj@ucw.cz>
 *	(c) 2005 Tomas Valla <tom@ucw.cz>
 */

#ifndef _LESTRADE_INTERNAL_H
#define _LESTRADE_INTERNAL_H

#include "lestrade/lestrade.h"

#include <setjmp.h>

/* Debugging */

#define DUMP(x...) fprintf(stderr, x)

/* Limits */

#define HARD_MAX_WORDS 32			/* Must be at most 32 (bits per int) */
#define HARD_MAX_PHRASES 8
#define HARD_MAX_NEARS 8
#define HARD_MAX_BOOLS 32			/* Must be at most 32 (bits per int) */
#define HARD_MAX_OCCURS 3			/* Must be at most 7 (flags in match-full.c) */
#define HARD_MAX_PHRASE_LEN 8

/* init.c: Configuration */

extern uns lex_max_gap, lex_max_ctrl_len, lex_max_hex_len, lex_max_len;
extern uns lex_min_len_ign, lex_min_len;
extern uns max_words, max_phrases, max_nears, max_bools, max_occurs, max_phrase_len;
extern uns accent_penalty;
extern uns morph_lemma_penalty, morph_basic_penalty, morph_other_penalty, morph_preacc_penalty;
extern uns default_weight, word_weight_scale;
extern uns near_bonus_word, near_penalty_gap, near_bonus_connect;
extern uns wild_min_prefix, wild_max_variants, wild_silent_fail, wild_max_cached;

enum word_class {
  WC_NORMAL,					/* Just an ordinary word */
  WC_IGNORED, 					/* Silently ignored */
  WC_GARBAGE,					/* Takes place (word number), but isn't indexed */
  WC_BREAK,					/* Explicit sentence break */
  WC_COMPLEX, 					/* [Not used by Lestrade] */
  WC_CONTEXT,					/* [Not used by Lestrade] */
};

struct exception {
  cnode n;
  enum word_class class;
  byte *w;
};

extern clist lex_exceptions;
extern uns lestrade_initialized;

/* query.c and internal representation of queries */

struct query {
  struct lestrade_query *Q;			/* Public part of the query */
  struct mempool *pool;
  jmp_buf err_jmp;
  struct lestrade_expr *prep_expr;			/* Preprocessed query expression */
  uns num_words;					/* Words, phrases and nears */
  struct word *words[HARD_MAX_WORDS];
  uns num_phrases;
  struct phrase *phrases[HARD_MAX_PHRASES];
  uns num_nears;
  struct phrase *nears[HARD_MAX_NEARS];
  uns num_bool_ids;					/* Boolean matching */
  struct word *words_by_bool_id[HARD_MAX_BOOLS];
  struct phrase *phrases_by_bool_id[HARD_MAX_BOOLS];
  u32 *bool_map;
  uns bool_bytes;
  struct trail_entry *trail, *trail_pos, *trail_end;	/* The word match trail */
  uns trail_size;
  uns trail_items;
  uns trail_sorted;
  uns trail_word_mask;
  uns trail_bool_mask;
  struct lestrade_match *matches, *matches_pos, *matches_end;	/* The match list */
  uns matches_size;
  uns matches_items;
  struct result_heap_entry *result_heap;		/* The result heap */
  struct lestrade_result *result_notes;
  uns result_cnt;
  uns result_max;
  struct detail_status *details;			/* Status of match_detail */
  struct ref_chain *ref_chains;				/* Reference chains found in the index */
  uns ref_cnt;
  uns ref_max;
  struct ref_chain **ref_heap;
  uns ref_heap_size;
};

struct lestrade_expr *expr_new(struct query *q, uns type);
struct lestrade_expr *expr_dup(struct query *q, struct lestrade_expr *x);
struct lestrade_expr *expr_new_op(struct query *q, uns type, struct lestrade_expr *l, struct lestrade_expr *r);
void expr_dump(struct lestrade_expr *x);

void query_flag_used(struct query *q, struct lestrade_expr *x);
struct lestrade_expr *query_simplify(struct query *q, struct lestrade_expr *x);

void NONRET query_error(struct query *q, int err, byte *msg, ...);

static inline int
query_tracing(struct query *q, uns level)
{
  return q->Q->trace_mask & level;
}

void query_trace(struct query *q, uns level, byte *msg, ...);

int bool_prepare(struct query *q, struct lestrade_expr *x);
void bool_cleanup(struct query *q);

/* index.c */

struct lestrade_index {
  byte name[64];
  byte *map_start;
  uns map_length;
  struct index_header *hdr;
  struct custom_index_hdr *cih;
  struct card_attrs *attrs;
  u32 *lexicon_ptrs;
  u32 *lexicon_zones;
  byte *lexicon;
  byte *refs;
};

/* words.c */

struct word {
  cnode n;					/* Node in atom->words */
  uns id;
  uns pos;					/* Word position in atom->words as told by the lexmapper */
  int bool_id;
  int best_weight;
  uns word_class;				/* WC_xxx */
  uns first_wild;				/* Position of the first `*' or `?' wildcard, 0 if none present */
  uns is_used;					/* Referenced by the boolean expression */
  uns is_accented;				/* Word contains accents */
  u32 type_set;					/* Set of WT_xxx types allowed */
  int weight;
  uns not_matchable;				/* Cannot be matched in current index */
  clist variants;
  uns var_cnt;
  byte *w;
  byte *unaccented_w;
  struct wildpatt *wild, *wild_unacc;		/* Compiled wildcard matching patterns */
};

struct variant {
  cnode n;
  byte penalty;
  byte flags;
  byte ignore_accents;
  byte *w;
};

enum var_flags {
  VF_WILD = 1,					/* Contains wildcards */
  VF_MORPH = 2,					/* Generated by morphological expansion */
  VF_LEMMA = 4,					/* This entry is a lemma */
  VF_BASIC = 8,					/* This entry is a basic form */
  VF_PREACCENT = 16,				/* Needed to modify accents of the source for to get this morphological variant */
};

struct phrase {
  u32 word_mask;				/* Set of word ID's used in this phrase */
  uns relpos[HARD_MAX_PHRASE_LEN];		/* Relative positions of the words */
  byte word_to_idx[HARD_MAX_WORDS];		/* Maps word id to position in phrase, offset by 1 */
  byte next_same_word[HARD_MAX_PHRASE_LEN];	/* Next occurence of the same word, offset by 1 */
  uns length;					/* Number of words */
  int weight;
  uns bool_id;
  struct word *words[HARD_MAX_WORDS];		/* Words from which the phrase consists */
  uns is_used;					/* Referenced by the boolean expression */
};

void words_global_init(void);
struct lestrade_expr *words_expand(struct query *q, struct lestrade_expr *x);
void word_dump(struct word *w);
void word_dump_full(struct word *w);
void phrase_dump(struct phrase *p);
byte *format_flags(byte *buf, byte *patt, uns flags);

/* match*.c (only basic functions; the rest is in match.h) */

void trail_cleanup(struct query *q);
void matches_cleanup(struct query *q);
void full_global_init(void);
void details_global_init(void);
void details_cleanup(struct query *q);
void index_cleanup(struct query *q);

/* results.c */

void results_init(struct query *q);
void results_cleanup(struct query *q);
void results_add(struct query *q, struct card_attrs *ca, int qq, struct custom_result *cr);

#endif
