/*
 *	The Lestrade Search Library -- Fulltext Matching
 *
 *	(c) 2005 Martin Mares <mj@ucw.cz>
 */

#undef LOCAL_DEBUG

#include "ucw/lib.h"
#include "ucw/fastbuf.h"
#include "ucw/unicode.h"
#include "ucw/unaligned.h"
#include "ucw/mempool.h"
#include "ucw/wildmatch.h"
#include "ucw/prime.h"
#include "charset/unicat.h"
#include "lestrade/internal.h"
#include "lestrade/tagged-text.h"
#include "lestrade/match.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

/* We use a heavily modified version of the lexhash module */

struct query *fh_query;
static struct verbum **fh_hash_table;
static struct mempool *fh_pool;
static uns fh_hash_size = 256;
static uns fh_hash_count;
static uns fh_hash_limit;
uns full_cnt;

static clist fw_wild_words;
static uns fw_have_wild_words, fw_min_prefix;
static uns fw_cached_words, fw_seen_words, fw_cache_misses;

#define FH_WALK(var) \
  for (uns lwi=0; lwi<fh_hash_size; lwi++) \
    for (struct verbum *var=fh_hash_table[lwi]; var; var=var->next)

/* The number of bits the hash is rotated by after every word.
 * It should be prime with the word size.  */
#define SHIFT_BITS 5

static inline uns PURE
fh_wlen(byte *x)
{
  uns l = 1;

  while (x[3])
    {
      x += 4;
      l++;
    }
  return l;
}

static inline u32 PURE
fh_hash(u32 *x, uns l)
{
  u32 h = *x++;
  while (--l)
    h = ROL(h, SHIFT_BITS) ^ *x++;
  return h % fh_hash_size;
}

static void
fh_alloc_hash(void)
{
  fh_hash_size = nextprime(fh_hash_size);
  fh_hash_limit = fh_hash_size*2;
  fh_hash_table = mp_alloc_zero(fh_pool, sizeof(struct verbum *) * fh_hash_size);
  fh_hash_count = 0;
}

static void
fh_rehash(uns newsize)
{
  struct verbum **oht = fh_hash_table;
  uns i, h;
  uns on = fh_hash_size;
  struct verbum *v, *next, **nht;

  fh_hash_size = newsize;
  fh_alloc_hash();
  nht = fh_hash_table;
  for (i=0; i<on; i++)
    for (v=oht[i]; v; v=next)
      {
	next = v->next;
	h = fh_hash((u32*)v->word, fh_wlen(v->word));
	v->next = nht[h];
	nht[h] = v;
      }
}

static inline struct verbum *
fh_hash_new(u32 *x, uns l, uns h)
{
  struct verbum *v = mp_alloc_fast(fh_pool, sizeof(struct verbum) + 4*l);
  struct verbum **w = &fh_hash_table[h];
  uns i;

  while (*w)
    w = &(*w)->next;
  v->next = NULL;
  *w = v;
  v->class = WC_NORMAL;
  v->num_occurs = 0;
  v->has_accented_variants = 0;
  v->unaccented = NULL;
  for (i=0; i<l; i++)
    ((u32*)v->word)[i] = x[i];
  if (++fh_hash_count > fh_hash_limit)
    fh_rehash(2*fh_hash_size);
  return v;
}

static inline struct verbum *
fh_do_lookup(byte *ww, byte *w)
{
  uns l, h, i;
  struct verbum *v;

#if CPU_ALLOW_UNALIGNED
  *(u32 *)w = 0;
#else
  switch ((w - ww) & 3)
    {
    default: *w++ = 0;
    case 1:  *w++ = 0;
    case 2:  *w++ = 0;
    case 3:  *w   = 0;
    }
#endif
  l = (w - ww + 4)/4;
  h = fh_hash((u32*)ww, l);
  for (v=fh_hash_table[h]; v; v=v->next)
    {
      for (i=0; i<l; i++)
	if (((u32*)v->word)[i] != ((u32*)ww)[i])
	  goto cont;
      return v;
    cont: ;
    }

  return NULL;
}

static struct verbum *fw_lookup_wild(byte *w, byte *wend, uns ulen, uns nonacc);

inline struct verbum *
fh_lookup(u16 *uni, uns ulen)
{
  byte ww[MAX_WORD_BYTES+4], *w;
  uns u;

  w = ww;
  while (ulen--)
    {
      u = *uni++;
      u = Utolower(u);
      w = utf8_put(w, u);
    }
  struct verbum *v = fh_do_lookup(ww, w);
  if (!v && fw_have_wild_words)
    v = fw_lookup_wild(ww, w, ulen, 0);
  return v;
}

inline struct verbum *
fh_lookup_nonacc(u16 *uni, uns ulen, uns *p_seen_accents)
{
  byte ww[MAX_WORD_BYTES+4], *w;
  uns u, uu;
  uns seen_accents = 0;

  w = ww;
  while (ulen--)
    {
      u = *uni++;
      u = Utolower(u);
      uu = Uunaccent(u);
      if (u != uu)
	seen_accents = 1;
      w = utf8_put(w, uu);
    }
  *p_seen_accents = seen_accents;
  struct verbum *v = fh_do_lookup(ww, w);
  if (!v && fw_have_wild_words)
    v = fw_lookup_wild(ww, w, ulen, seen_accents);
  return v;
}

static struct verbum *
fh_insert(byte *c, uns noacc, uns *created_p)
{
  byte ww[MAX_WORD_BYTES+1];
  u32 *ww32 = (u32*)ww;
  byte *w = ww;
  uns u, l, h, ulen=0, seen_accents=0;
  struct verbum *v;
  while (*c)
    {
      c = utf8_get(c, &u);
      u = Utolower(u);
      if (noacc)
	{
	  uns u0 = u;
	  u = Uunaccent(u);
	  if (u != u0)
	    seen_accents=1;
	}
      w = utf8_put(w, u);
      ulen++;
    }
  PUT_U32(w, 0);
  l = (w-ww+4)/4;
  h = fh_hash(ww32, l);
  for (v=fh_hash_table[h]; v; v=v->next)
    if (!strcmp(v->word, ww))
      {
	*created_p = 0;
	return v;
      }
  *created_p = 1;
  if (!fw_have_wild_words || !(v = fw_lookup_wild(ww, w, ulen, seen_accents)))
    v = fh_hash_new(ww32, l, h);
  return v;
}

static struct verbum *
fh_raw_insert(byte *ww, byte *w)
{
  u32 *ww32 = (u32*)ww;
  uns l = (w-ww+4)/4;
  uns h = fh_hash(ww32, l);
  return fh_hash_new(ww32, l, h);
}

static void
fh_load_exceptions(void)
{
  struct exception *e;

  CLIST_WALK(e, lex_exceptions)
    {
      uns created_p = 0;
      struct verbum *v = fh_insert(e->w, 1, &created_p);
      if (!created_p)
	die("Lexical exception <%s> defined twice", e->w);
      v->class = e->class;
    }
}

static void
full_dump_words(struct query *q)
{
  if (query_tracing(q, TRACE_FULLTEXT))
    {
      msg(L_DEBUG, "Running fulltext matcher with the following lexicon:");
      FH_WALK(v)
	{
	  DUMP("\t<%s> class=%d hasacc=%d", v->word, v->class, v->has_accented_variants);
	  for (uns i=0; i<v->num_occurs; i++)
	    DUMP(" (wrd=%d,pen=%d/%d)", v->word_idx[i], v->penalty[i], v->penalty_nonacc[i]);
	  DUMP("\n");
	}
    }
}

static void
fh_add_occur(struct query *q, struct verbum *v, uns word_idx, uns penalty, uns penalty_nonacc)
{
  DBG("Adding occurence <%s> w=%d p=%d/%d", v->word, word_idx, penalty, penalty_nonacc);

  for (uns i=0; i<v->num_occurs; i++)
    if (v->word_idx[i] == word_idx)
      {
	v->penalty[i] = MIN(v->penalty[i], penalty);
	v->penalty_nonacc[i] = MIN(v->penalty_nonacc[i], penalty_nonacc);
	return;
      }

  if (v->num_occurs >= max_occurs)
    query_error(q, LESTRADE_ERR_TOO_MANY_OCCURENCES, "Too many occurences of a single word");
  uns i = v->num_occurs++;
  v->word_idx[i] = word_idx;
  v->penalty[i] = penalty;
  v->penalty_nonacc[i] = penalty_nonacc;
}

static void
full_fold_words(struct query *q)
{
  FH_WALK(va)
    {
      struct verbum *vu = va->unaccented;
      if (vu)
	{
	  for (uns i=0; i<vu->num_occurs; i++)
	    if (vu->penalty_nonacc[i] < 255)
	      {
		uns va_occ = va->num_occurs;
		uns j;
		for (j=0; j<va_occ; j++)
		  if (va->word_idx[j] == va->word_idx[i])
		    break;
		if (j < va_occ)
		  {
		    if (va->penalty[j] > vu->penalty_nonacc[i])
		      va->penalty[j] = va->penalty_nonacc[i];
		  }
		else
		  fh_add_occur(q, va, vu->word_idx[i], vu->penalty_nonacc[i], 255);
	      }
	}
    }
}


struct wild_word {
  cnode n;
  struct word *w;
  struct variant *first_var;
};

#define FW_HASH_SIZE 8192
#define FW_HASH_BYTES ((FW_HASH_SIZE+7)/8)
static byte fw_hash[FW_HASH_BYTES];

static void
fw_init(void)
{
  fw_have_wild_words = 0;
  fw_min_prefix = ~0U;
  clist_init(&fw_wild_words);
}

static void
fw_add(struct word *w)
{
  struct wild_word *ww = mp_alloc(fh_query->pool, sizeof(*ww));
  ww->w = w;
  ww->first_var = (struct variant *) clist_head(&w->variants);
  clist_add_tail(&fw_wild_words, &ww->n);
  fw_have_wild_words++;
  fw_min_prefix = MIN(fw_min_prefix, w->first_wild);
}

static uns
fw_prefix_hash(byte *w)
{
  uns len = fw_min_prefix;
  uns u, h = 0;
  while (len--)
    {
      w = (byte*)utf8_get(w, &u);
      h = 37*h + u;
    }
  return h % FW_HASH_SIZE;
}

static void
fw_seal(void)
{
  DBG("Constructing prefix hash for length %d", fw_min_prefix);
  bzero(fw_hash, sizeof(fw_hash));
  CLIST_FOR_EACH(struct wild_word *, ww, fw_wild_words)
    {
      struct word *w = ww->w;
      uns h;
      h = fw_prefix_hash(w->unaccented_w);
      fw_hash[h/8] |= (1 << (h%8));
      h = fw_prefix_hash(w->w);
      fw_hash[h/8] |= (1 << (h%8));
    }
  fw_cached_words = 0;
  fw_cache_misses = 0;
  fw_seen_words = 0;
}

static struct verbum *
fw_lookup_wild(byte *w, byte *wend, uns ulen, uns nonacc)
{
  if (ulen < fw_min_prefix)
    return NULL;
  uns h = fw_prefix_hash(w);
  if (!(fw_hash[h/8] & (1 << (h%8))))
    return NULL;

  DBG("Wildcard lookup for <%s>", w);
  struct verbum *bum = NULL;
  fw_seen_words++;
  CLIST_FOR_EACH(struct wild_word *, ww, fw_wild_words)
    {
      struct word *wo = ww->w;
      int match = -1;
      if (nonacc)
	{
	  if (wp_match(wo->wild_unacc, w))
	    {
	      match = ww->first_var->penalty;
	      if (!wo->is_accented)
		match += accent_penalty;
	    }
	}
      else
	{
	  if (wp_match(wo->wild, w))
	    match = ww->first_var->penalty;
	}
      if (match >= 0)
	{
	  DBG("\tMatched <%s> penalty %d", wo->w, match);
	  if (!bum)
	    bum = fh_raw_insert(w, wend);
	  fh_add_occur(fh_query, bum, wo->id, match, ww->first_var->penalty + accent_penalty);
	  if (nonacc && wo->is_accented)
	    bum->has_accented_variants = 1;
	}
    }
  if (!bum)
    {
      if (fw_cached_words < wild_max_cached)
	{
	  fw_cached_words++;
	  bum = fh_raw_insert(w, wend);
	  DBG("\tMismatch cached");
	}
      else
	{
	  fw_cache_misses++;
	  DBG("\tMismatch not cached");
	}
    }
  return bum;
}

static void
full_init_words(struct query *q)
{
  fh_query = q;
  fh_pool = q->pool;
  fh_alloc_hash();
  fh_load_exceptions();
  fw_init();

  /* Process all wildcard words first */
  for (uns i=0; i<q->num_words; i++)
    {
      struct word *w = q->words[i];
      if (w->first_wild)
	{
	  w->not_matchable = 0;
	  fw_add(w);
	}
    }
  fw_seal();

  /* And then the normal ones */
  for (uns i=0; i<q->num_words; i++)
    {
      struct word *w = q->words[i];
      if (w->word_class == WC_NORMAL && !w->first_wild)
	{
	  w->not_matchable = 0;
	  CLIST_FOR_EACH(struct variant *, v, w->variants)
	    {
	      uns va_created_p, vu_created_p;
	      struct verbum *va = fh_insert(v->w, 0, &va_created_p);
	      if (va_created_p)
		{
		  struct verbum *vu = fh_insert(v->w, 1, &vu_created_p);
		  if (va != vu)
		    {
		      va->class = vu->class;
		      va->unaccented = vu;
		      vu->has_accented_variants = 1;
		    }
		}
	      if (v->ignore_accents)
		{
		  struct verbum *vu = va->unaccented;
		  if (vu)
		    fh_add_occur(q, vu, w->id, v->penalty + accent_penalty, v->penalty + accent_penalty);
		  else
		    fh_add_occur(q, va, w->id, v->penalty, v->penalty + accent_penalty);
		}
	      else
		fh_add_occur(q, va, w->id, v->penalty, 255);
	    }
	}
    }

  full_fold_words(q);
  full_dump_words(q);
}

typedef struct verbum *word_id_t;
#define LM_MAP_ARGS struct fastbuf *scan_char_fb
#define LM_SCAN_CHAR(u) ( ((u)=bget_tagged_char(scan_char_fb)) == ~0U ? 0 : 1)

static enum word_class lm_lookup(enum word_class orig_class, u16 *uni, uns ulen, word_id_t *idp)
{
  if (orig_class != WC_NORMAL)
    return orig_class;

  uns seen_accents;
  struct verbum *v = fh_lookup_nonacc(uni, ulen, &seen_accents);
  if (v)
    {
      if (!seen_accents)
	*idp = (word_id_t)((uintptr_t)v | 1);	/* Yes, it's a hack, but I need to store that bit somewhere :) */
      else
	{
	  *idp = v;
	  if (v->has_accented_variants)
	    {
	      struct verbum *va = fh_lookup(uni, ulen);
	      if (va)
		{
		  *idp = (word_id_t)((uintptr_t)va | 1);
		  v = va;
		}
	    }
	}
      return v->class;
    }
  else
    return (ulen < lex_min_len_ign) ? WC_IGNORED : (ulen < lex_min_len) ? WC_GARBAGE : WC_NORMAL;
}

static inline void lm_got_word(uns pos, uns cat, word_id_t w)
{
  if (!w)
    return;
  struct verbum *v = (word_id_t)((uintptr_t)w & ~(uintptr_t)1);
  uns precise = (uintptr_t)w & 1;

  DBG("@%d: <%s> prec=%d cat=%d", pos, v->word, precise, cat);
  for (uns i=0; i<v->num_occurs; i++)
    {
      uns pen = (precise ? v->penalty : v->penalty_nonacc) [i];
      if (pen < 255)
	trail_add(fh_query, pos, v->word_idx[i], cat, pen);
    }
}

static inline void lm_got_complex(uns pos UNUSED, uns cat UNUSED, word_id_t root UNUSED, word_id_t w UNUSED, uns dir UNUSED)
{
  ASSERT(0);
}

#include "lestrade/lexmap.h"

static int
full_card(struct query *q, struct card_attrs *ca, struct fastbuf *fb)
{
  DBG("--- SNIP ---");
  ASSERT(fh_query && q->bool_map);
  full_cnt++;
  lm_doc_start();
  trail_init(q, 1);
  lm_map_text(fb);
  bclose(fb);
  int qq = match_single(q, ca);
  return (qq > -INFTY);
}

int
full_init(struct query *q)
{
  full_cnt = 0;
  if (!setjmp(q->err_jmp))
    {
      struct lestrade_expr *x = q->prep_expr;
      full_init_words(q);
      bool_prepare(q, x);
    }
  return q->Q->error_code;
}

int
full_cleanup(struct query *q)
{
  if (fw_have_wild_words)
    query_trace(q, TRACE_GENERAL, "Fulltext matcher: %d words inspected for wildcards, %d cache misses", fw_seen_words, fw_cache_misses);
  bool_cleanup(q);
  fh_query = NULL;
  return 0;
}

int
lestrade_eval_cards_init(struct lestrade_query *Q)
{
  struct query *q = Q->q;
  query_trace(q, TRACE_GENERAL, "Fulltext matcher: initializing");
  return full_init(q);
}

int
lestrade_eval_cards_cleanup(struct lestrade_query *Q)
{
  struct query *q = Q->q;
  int rc = full_cleanup(q);
  query_trace(q, TRACE_GENERAL, "Fulltext matcher: processed %d cards", full_cnt);
  return rc;
}

int
lestrade_eval_card(struct lestrade_query *Q, struct card_attrs *attrs)
{
  return full_card(Q->q, attrs, Q->fetcher(attrs, Q->fetcher_priv));
}

int
lestrade_eval_cards(struct lestrade_query *Q, card_enumerator_t source, void *enum_priv)
{
  struct query *q = Q->q;
  struct fastbuf *fb;
  struct card_attrs ca;
  int cnt = 0;
  while (fb = source(&ca, enum_priv))
    cnt += full_card(q, &ca, fb);
  return cnt;
}

void
full_global_init(void)
{
  lm_init();
}
