/*
 *	The Lestrade Search Library -- Operations with Queries
 *
 *	(c) 2005 Martin Mares <mj@ucw.cz>
 */

#include "ucw/lib.h"
#include "ucw/mempool.h"
#include "lestrade/internal.h"

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

/* Tracing and reporting of errors */

void
query_error(struct query *q, int err, byte *msg, ...)
{
  va_list args;
  va_start(args, msg);
  q->Q->error_code = err;
  int len = 1 + vsnprintf(NULL, 0, msg, args);
  q->Q->error_msg = mp_alloc(q->pool, len);
  vsprintf(q->Q->error_msg, msg, args);
  query_trace(q, TRACE_GENERAL, "ERROR: %s", q->Q->error_msg);
  va_end(args);
  longjmp(q->err_jmp, err);
}

void
query_trace(struct query *q, uns level, byte *msg, ...)
{
  if (query_tracing(q, level))
    {
      va_list args;
      va_start(args, msg);
      vmsg(L_DEBUG, msg, args);
      va_end(args);
    }
}

/* Manipulating expressions */

struct lestrade_expr *
expr_new(struct query *q, uns type)
{
  struct lestrade_expr *x = mp_alloc_zero(q->pool, sizeof(*x));
  x->type = type;
  return x;
}

struct lestrade_expr *
expr_new_op(struct query *q, uns type, struct lestrade_expr *l, struct lestrade_expr *r)
{
  struct lestrade_expr *e = expr_new(q, type);
  e->u.op.l = l;
  e->u.op.r = r;
  return e;
}

struct lestrade_expr *
expr_dup(struct query *q, struct lestrade_expr *x)
{
  struct lestrade_expr *y = mp_alloc(q->pool, sizeof(*x));
  *y = *x;
  return y;
}

static void
atom_dump(struct lestrade_atom *a)
{
  DUMP("(ATOM <%s> types=%x wt=%d sns=%d acc=%d mor=%d)",
       a->word, a->type_set, a->weight, a->sense, a->accent_mode, a->morph_mode);
}

static void
do_expr_dump(struct lestrade_expr *x, int level, int close)
{
  for (int i=0; i<level; i++)
    DUMP("\t");
  DUMP("(");
  level++;
  close++;
  switch (x->type)
    {
    case EX_MATCH:
      DUMP("MATCH");
      struct lestrade_atom *a;
      CLIST_WALK(a, x->u.match.atoms)
	{
	  DUMP("\n");
	  for (int i=0; i<level; i++)
	    DUMP("\t");
	  atom_dump(a);
	}
      break;
    case EX_AND:
    case EX_OR:
      DUMP("%s\n", (x->type == EX_AND ? "AND" : "OR"));
      do_expr_dump(x->u.op.l, level, 0);
      do_expr_dump(x->u.op.r, level, close);
      return;
    case EX_NOT:
      DUMP("NOT\n");
      do_expr_dump(x->u.op.l, level, close);
      return;
    case EX_ANY:
      DUMP("ANY");
      break;
    case EX_NONE:
      DUMP("NONE");
      break;
    case EX_IGNORE:
      DUMP("IGNORE");
      break;
    case EX_WORD: ;
      struct word *w = x->u.word;
      DUMP("WORD ");
      word_dump(w);
      break;
    case EX_PHRASE:
      DUMP("PHRASE ");
      phrase_dump(x->u.phrase);
      break;
    default:
      die("do_expr_dump: called for unknown type %d", x->type);
    }
  while (close--)
    DUMP(")");
  DUMP("\n");
}

void
expr_dump(struct lestrade_expr *x)
{
  do_expr_dump(x, 1, 0);
}

static struct lestrade_expr *
expr_cow(struct query *q, struct lestrade_expr *x, struct lestrade_expr *l, struct lestrade_expr *r)
{
  if (x->u.op.l == l && x->u.op.r == r)
    return x;
  else
    return expr_new_op(q, x->type, l, r);
}

/* Preprocessing of queries */

static struct lestrade_expr *
query_expand_words(struct query *q, struct lestrade_expr *x)
{
  switch (x->type)
    {
    case EX_MATCH:
      return words_expand(q, x);
    case EX_AND:
    case EX_OR:
    case EX_NOT:
      return expr_cow(q, x, query_expand_words(q, x->u.op.l), query_expand_words(q, x->u.op.r));
    case EX_ANY:
    case EX_NONE:
      return x;
    default:
      ASSERT(0);
    }
}

static void
query_flag_expr(struct query *q, struct lestrade_expr *x)
{
  if (!x)
    return;
  switch (x->type)
    {
    case EX_WORD:
      x->u.word->is_used = 1;
      break;
    case EX_PHRASE:
      x->u.phrase->is_used = 1;
      break;
    case EX_AND:
    case EX_OR:
    case EX_NOT:
      query_flag_expr(q, x->u.op.l);
      query_flag_expr(q, x->u.op.r);
      break;
    case EX_ANY:
    case EX_NONE:
      return;
    default:
      ASSERT(0);
    }
}

void
query_flag_used(struct query *q, struct lestrade_expr *x)
{
  for (uns i=0; i<q->num_words; i++)
    q->words[i]->is_used = 0;
  for (uns i=0; i<q->num_phrases; i++)
    q->phrases[i]->is_used = 0;
  query_flag_expr(q, x);
}

struct lestrade_expr *
query_simplify(struct query *q, struct lestrade_expr *x)
{
  switch (x->type)
    {
    case EX_WORD:
      if (x->u.word->not_matchable)
	return expr_new(q, EX_NONE);
      else
	return x;
    case EX_PHRASE:
      {
	struct phrase *p = x->u.phrase;
	for (uns i=0; i<p->length; i++)
	  if (p->words[i]->not_matchable)
	    return expr_new(q, EX_NONE);
	return x;
      }
    case EX_ANY:
    case EX_NONE:
      return x;
    case EX_AND:
      {
	struct lestrade_expr *l = query_simplify(q, x->u.op.l);
	struct lestrade_expr *r = query_simplify(q, x->u.op.r);
	if (l->type == EX_NONE)
	  return l;
	if (r->type == EX_NONE)
	  return r;
	if (l->type == EX_ANY)
	  return r;
	if (r->type == EX_ANY)
	  return l;
	return expr_cow(q, x, l, r);
      }
    case EX_OR:
      {
	struct lestrade_expr *l = query_simplify(q, x->u.op.l);
	struct lestrade_expr *r = query_simplify(q, x->u.op.r);
	if (l->type == EX_ANY)
	  return l;
	if (r->type == EX_ANY)
	  return r;
	if (l->type == EX_NONE)
	  return r;
	if (r->type == EX_NONE)
	  return l;
	return expr_cow(q, x, l, r);
      }
    case EX_NOT:
      {
	struct lestrade_expr *l = query_simplify(q, x->u.op.l);
	if (l->type == EX_ANY)
	  return expr_new(q, EX_NONE);
	if (l->type == EX_NONE)
	  return expr_new(q, EX_ANY);
	if (l->type == EX_NOT)
	  return l->u.op.l;
	return expr_cow(q, x, l, NULL);
      }
    default:
      ASSERT(0);
    }
}

static int
expr_matched_by_any(struct lestrade_expr *x)
{
  switch (x->type)
    {
    case EX_ANY:
      return 1;
    case EX_NONE:
      return -1;
    case EX_WORD:
    case EX_PHRASE:
      return -1;
    case EX_NOT:
      return -expr_matched_by_any(x->u.op.l);
    case EX_AND:
    case EX_OR:
      {
	int ll = expr_matched_by_any(x->u.op.l);
	int rr = expr_matched_by_any(x->u.op.r);
	return (x->type == EX_AND) ? MIN(ll, rr) : MAX(ll, rr);
      }
    default:
      ASSERT(0);
    }
}

/* Evaluation of boolean part of the expression */

static void
bool_make_map(struct query *q, struct lestrade_expr *x, u32 *Z)
{
  u32 *X, *Y, *X0, *Y0;
  uns c, i;

  switch (x->type)
    {
    case EX_AND:
    case EX_OR:
      X = X0 = xmalloc(q->bool_bytes);
      Y = Y0 = xmalloc(q->bool_bytes);
      bool_make_map(q, x->u.op.l, X);
      bool_make_map(q, x->u.op.r, Y);
      if (x->type == EX_AND)
	for(c=0; c<q->bool_bytes; c+=4)
	  *Z++ = *X++ & *Y++;
      else
	for(c=0; c<q->bool_bytes; c+=4)
	  *Z++ = *X++ | *Y++;
      xfree(Y0);
      xfree(X0);
      break;
    case EX_NOT:
      X = X0 = xmalloc(q->bool_bytes);
      bool_make_map(q, x->u.op.l, X);
      for(c=0; c<q->bool_bytes; c+=4)
	*Z++ = ~*X++;
      xfree(X0);
      break;
    case EX_PHRASE:
      i = x->u.phrase->bool_id;
      goto got_id;
    case EX_WORD:
      ASSERT(!x->u.word->not_matchable);
      i = x->u.word->bool_id;
    got_id:
      if (i < 5)			/* Fits in u32 */
	{
	  uns k = 1 << i;
	  uns t;
	  u32 j = 0;
	  for(t=0; t < 32; t++)
	    if (t & k)
	      j |= 1 << t;
	  for(c=0; c<q->bool_bytes; c+=4)
	    *Z++ = j;
	}
      else
	{
	  uns per = 1 << (i - 5);
	  uns t = 2*per*4;
	  for(c=0; c<q->bool_bytes; c+=t)
	    {
	      for(i=0; i<per; i++)
		*Z++ = 0;
	      for(i=0; i<per; i++)
		*Z++ = ~0;
	    }
	}
      break;
    case EX_ANY:
      memset(Z, 255, q->bool_bytes);
      break;
    case EX_NONE:
      memset(Z, 0, q->bool_bytes);
      break;
    default:
      ASSERT(0);
    }
}

int
bool_prepare(struct query *q, struct lestrade_expr *x)
{
  ASSERT(x);
  ASSERT(!q->bool_map);
  query_flag_used(q, x);

  q->num_bool_ids = 0;
  for (uns i=0; i<q->num_words; i++)
    {
      struct word *w = q->words[i];
      if (w->is_used)
	{
	  w->bool_id = q->num_bool_ids++;
	  q->words_by_bool_id[w->bool_id] = w;
	}
    }
  for (uns i=0; i<q->num_phrases; i++)
    {
      struct phrase *p = q->phrases[i];
      if (p->is_used)
	{
	  p->bool_id = q->num_bool_ids++;
	  q->phrases_by_bool_id[p->bool_id] = p;
	}
    }
  if (q->num_bool_ids > max_bools)
    query_error(q, LESTRADE_ERR_TOO_MANY_BOOLS, "Boolean expression too complex");

  if (q->num_bool_ids <= 5)
    q->bool_bytes = sizeof(u32);
  else
    q->bool_bytes = 1 << (q->num_bool_ids - 3);
  query_trace(q, TRACE_PREPARE, "Constructing boolean expression map of order %d and size %d", q->num_bool_ids, q->bool_bytes);
  q->bool_map = xmalloc(q->bool_bytes);
  bool_make_map(q, x, q->bool_map);

  if (q->bool_map[0] & 1)
    query_error(q, LESTRADE_ERR_ANY_DOCUMENT_MATCHES, "Any document matches");	/* This probably cannot happen */
  for (uns c=0; c<q->bool_bytes/4; c++)
    if (q->bool_map[c])
      return 1;
  return 0;
}

void
bool_cleanup(struct query *q)
{
  if (q->bool_map)
    {
      xfree(q->bool_map);
      q->bool_map = NULL;
    }
}

/* Public interface */

struct lestrade_query *
lestrade_new_query(void)
{
  struct mempool *mp = mp_new(4096);
  struct query *q = mp_alloc_zero(mp, sizeof(*q));
  struct lestrade_query *Q = mp_alloc_zero(mp, sizeof(*Q));

  q->Q = Q;
  q->pool = mp;

  Q->q = q;
  Q->pool = mp;
  Q->max_results = 100;

  return Q;
}

void
lestrade_free_query(struct lestrade_query *Q)
{
  struct query *q = Q->q;
  bool_cleanup(q);
  trail_cleanup(q);
  matches_cleanup(q);
  details_cleanup(q);
  index_cleanup(q);
  results_cleanup(q);
  mp_delete(Q->pool);
}

void
lestrade_dump_query(struct lestrade_query *Q)
{
  DUMP("Query %p:\n", Q);
  expr_dump(Q->expr);
}

struct lestrade_expr *
lestrade_new_op(struct lestrade_query *Q, uns op, struct lestrade_expr *l, struct lestrade_expr *r)
{
  return expr_new_op(Q->q, op, l, r);
}

struct lestrade_expr *
lestrade_new_match(struct lestrade_query *Q)
{
  struct lestrade_expr *e = expr_new(Q->q, EX_MATCH);
  clist_init(&e->u.match.atoms);
  return e;
}

struct lestrade_atom *lestrade_add_atom(struct lestrade_query *Q, struct lestrade_expr *e, byte *word)
{
  struct lestrade_atom *a = mp_alloc_zero(Q->pool, sizeof(*a));
  clist_add_tail(&e->u.match.atoms, &a->n);
  a->word = mp_strdup(Q->pool, word);
  a->type_set = ~0U;
  a->weight = default_weight;
  a->sense = 1;
  a->accent_mode = ACCENT_MODE_STRICT;
  a->morph_mode = MORPH_MODE_OFF;
  clist_init(&a->words);
  return a;
}

static void
query_trace_dump(struct query *q, uns level, struct lestrade_expr *x, byte *desc)
{
  if (query_tracing(q, level))
    {
      query_trace(q, level, "%s:", desc);
      expr_dump(x);
    }
}

static void
query_trace_words(struct query *q, uns level)
{
  if (query_tracing(q, level))
    {
      msg(L_DEBUG, "Known words, phrases and nears:");
      for (uns i=0; i<q->num_words; i++)
	{
	  DUMP("\tWord %d: ", i);
	  word_dump_full(q->words[i]);
	}
      for (uns i=0; i<q->num_phrases; i++)
	{
	  DUMP("\tPhrase %d: ", i);
	  phrase_dump(q->phrases[i]);
	  DUMP("\n");
	}
      for (uns i=0; i<q->num_nears; i++)
	{
	  DUMP("\tNear %d: ", i);
	  phrase_dump(q->nears[i]);
	  DUMP("\n");
	}
    }
}

int
lestrade_init_query(struct lestrade_query *Q)
{
  struct query *q = Q->q;

  ASSERT(lestrade_initialized);
  if (!setjmp(q->err_jmp))
    {
      struct lestrade_expr *x = Q->expr;
      query_trace_dump(q, TRACE_GENERAL, x, "Initializing query");
      x = query_expand_words(q, x);
      query_trace_dump(q, TRACE_PREPARE, x, "After expansion of words");
      x = query_simplify(q, x);
      query_trace_dump(q, TRACE_PREPARE, x, "After simplification");
      if (expr_matched_by_any(x) > 0)
	query_error(q, LESTRADE_ERR_ANY_DOCUMENT_MATCHES, "Any document matches");
      query_trace_words(q, TRACE_PREPARE);
      q->prep_expr = x;
      results_init(q);
    }
  return Q->error_code;
}
