/*
 *	The Lestrade Search Library -- Processing of Words
 *
 *	(c) 2005 Martin Mares <mj@ucw.cz>
 */

#include "ucw/lib.h"
#include "ucw/mempool.h"
#include "ucw/unicode.h"
#include "ucw/wildmatch.h"
#include "lang/lang.h"
#include "lestrade/internal.h"
#include "charset/unicat.h"

#include <stdio.h>
#include <string.h>

static struct query *word_current_query;
static struct lestrade_atom *word_current_atom;
typedef struct word *word_id_t;

static struct lestrade_expr *
simple_fold(struct query *q, struct lestrade_expr *x)
{
  struct lestrade_atom *a;
  struct lestrade_expr *maybes, *sharps;
  uns seen_yes = 0;

  /* Process MAYBE parts of the query */
  maybes = NULL;
  CLIST_WALK(a, x->u.match.atoms)
    {
      ASSERT(a->xlat);
      if (!a->sense && a->xlat->type != EX_IGNORE)
	maybes = maybes ? expr_new_op(q, EX_OR, maybes, a->xlat) : a->xlat;
    }

  /* Process YES and NO parts */
  sharps = NULL;
  CLIST_WALK(a, x->u.match.atoms)
    if (a->sense && a->xlat->type != EX_IGNORE)
      {
	struct lestrade_expr *f = (a->sense > 0) ? a->xlat : expr_new_op(q, EX_NOT, a->xlat, NULL);
	sharps = sharps ? expr_new_op(q, EX_AND, sharps, f) : f;
	if (a->sense > 0)
	  seen_yes++;
      }

  if (!maybes && !sharps)
    query_error(q, LESTRADE_ERR_ONLY_NONIDX, "Simple search expression contains only non-indexed words");
  if (!maybes || !sharps)
    return maybes ? : sharps;
  if (seen_yes)
    maybes = expr_new_op(q, EX_OR, expr_new(q, EX_ANY), maybes);
  return expr_new_op(q, EX_AND, maybes, sharps);
}

static struct word *
word_alloc(struct query *q, struct lestrade_atom *a, byte *acc, byte *unacc)
{
  struct word *w = mp_alloc_zero(q->pool, sizeof(*w));
  if (q->num_words >= max_words)
    query_error(q, LESTRADE_ERR_TOO_MANY_WORDS, "Too many words");
  q->words[q->num_words] = w;
  w->id = q->num_words++;
  w->bool_id = -1;

  w->pos = ~0U;
  w->type_set = a->type_set;
  w->weight = a->weight;
  clist_init(&w->variants);
  w->w = mp_strdup(q->pool, acc);
  w->unaccented_w = mp_strdup(q->pool, unacc);
  w->is_accented = strcmp(acc, unacc);
  return w;
}

static void
word_classify(struct word *w, enum word_class orig_class, uns ulen)
{
  int c = -1;

  if (orig_class != WC_NORMAL)
    c = orig_class;
  else if (strchr(w->w, '*') || strchr(w->w, '?'))
    {
      for (w->first_wild = 0; w->w[w->first_wild] != '*' && w->w[w->first_wild] != '?'; w->first_wild++)
	;
      if (w->first_wild < wild_min_prefix)
	query_error(word_current_query, LESTRADE_ERR_WILD_TOO_BROAD, "Wildcard pattern too broad");
      c = WC_NORMAL;
    }
  else
    {
      CLIST_FOR_EACH(struct exception *, ex, lex_exceptions)
	if (!strcmp(ex->w, w->w))
	  {
	    c = ex->class;
	    break;
	  }
      if (c < 0)
	{
	  if (ulen < lex_min_len_ign)
	    c = WC_IGNORED;
	  else if (ulen < lex_min_len)
	    c = WC_GARBAGE;
	  else
	    c = WC_NORMAL;
	}
    }
  w->word_class = c;
}

static enum word_class
lm_lookup(enum word_class orig_class, u16 *uni, uns ulen, word_id_t *thisw)
{
  struct word *w;
  byte wbuf[MAX_WORD_BYTES+1], *wp=wbuf;
  byte ubuf[MAX_WORD_BYTES+1], *up=ubuf;

  if (!uni)
    return orig_class;
  for (uns i=0; i<ulen; i++)
    {
      uns u = uni[i];
      u = Utolower(u);
      wp = utf8_put(wp, u);
      u = Uunaccent(u);
      up = utf8_put(up, u);
    }
  *wp = *up = 0;
  w = word_alloc(word_current_query, word_current_atom, wbuf, ubuf);
  word_classify(w, orig_class, ulen);
  *thisw = w;
  return w->word_class;
}

static void
lm_got_word(uns pos, uns cat UNUSED, word_id_t p)
{
  if (p->pos != ~0U)
    clist_remove(&p->n);
  clist_add_tail(&word_current_atom->words, &p->n);
  p->pos = pos;
}

static inline void
lm_got_complex(uns pos UNUSED, uns cat UNUSED, word_id_t root UNUSED, word_id_t context UNUSED, uns dir UNUSED)
{
  ASSERT(0);
}

#define LM_SEARCH
#define LM_SCAN_CHAR(u) (text = utf8_get(text, &u), u)
#define LM_MAP_ARGS byte *text
#include "lestrade/lexmap.h"

byte *
format_flags(byte *buf, byte *patt, uns flags)
{
  uns l = strlen(patt) - 1;
  strcpy(buf, patt);
  for (uns i=0; i<=l; i++)
    if (!(flags & (1 << i)))
      buf[l-i] = '-';
  return buf;
}

void
word_dump(struct word *w)
{
  DUMP("<%s> unacc=<%s> pos=%d cls=%d wild=%d accented=%d types=%08x wt=%d vars=%d bool_id=%d",
       w->w, w->unaccented_w, w->pos, w->word_class, w->first_wild, w->is_accented, w->type_set, w->weight, w->var_cnt, w->bool_id);
}

void
phrase_dump(struct phrase *p)
{
  for (uns i=0; i<p->length; i++)
    DUMP("<%s>@+%d ", p->words[i]->w, p->relpos[i]);
  DUMP("wt=%d bool_id=%d word_mask=%x", p->weight, p->bool_id, p->word_mask);
}

static void
word_dump_var(struct variant *v)
{
  byte buf[9];
  DUMP("pen=%3d flg=%s ign_acc=%d  <%s>\n", v->penalty, format_flags(buf, "???PBLMW", v->flags), v->ignore_accents, v->w);
}

void
word_dump_full(struct word *w)
{
  word_dump(w);
  DUMP("\n");
  struct variant *v;
  CLIST_WALK(v, w->variants)
    {
      DUMP("\t\t");
      word_dump_var(v);
    }
}

static void
words_dump_atom(struct query *q, struct lestrade_atom *a, byte *when)
{
  if (query_tracing(q, TRACE_WORDS))
    {
      DUMP("%s:\n", when);
      struct word *w;
      CLIST_WALK(w, a->words)
	{
	  DUMP("\t");
	  word_dump_full(w);
	}
    }
}

static struct variant *
var_alloc(struct query *q, struct word *w, byte *t)
{
  struct variant *v = mp_alloc(q->pool, sizeof(*v));
  clist_add_tail(&w->variants, &v->n);
  v->penalty = 0;
  v->flags = (w->first_wild ? VF_WILD : 0);
  v->ignore_accents = 0;
  v->w = t;
  return v;
}

static void
word_expand(struct query *q, struct lestrade_atom *a, struct word *w)
{
  struct variant *v = var_alloc(q, w, w->w);
  if (a->accent_mode == ACCENT_MODE_STRIP)
    v->ignore_accents = 1;

  if (w->first_wild)
    {
      w->wild_unacc = w->wild = wp_compile(w->w, q->pool);
      if (a->accent_mode == ACCENT_MODE_STRIP && w->is_accented)
	w->wild_unacc = wp_compile(w->unaccented_w, q->pool);
      if (!w->wild || !w->wild_unacc)
	query_error(q, LESTRADE_ERR_WILD_TOO_LONG, "Wildcard pattern too long");
    }

#ifdef CONFIG_LANG
  else if (a->morph_mode > MORPH_MODE_OFF)
    CLIST_FOR_EACH(struct stemmer *, st, stemmer_list)
      {
	struct word_node req = {
	  .word_form = WORD_FORM_OTHER,
	  .stem_form = WORD_FORM_OTHER,
	  .unaccented = (a->morph_mode == MORPH_MODE_PREACCENT && a->accent_mode == ACCENT_MODE_STRIP),
	  .w = w->w
	};
	clist *res = lang_stem(st, &req, q->pool);
	if (res)
	  CLIST_FOR_EACH(struct word_node *, n, *res)
	  {
	    v = var_alloc(q, w, n->w);
	    v->flags |= VF_MORPH;
	    v->flags |= ((uns[]) { [WORD_FORM_LEMMA] = VF_LEMMA,
				   [WORD_FORM_BASIC] = VF_BASIC,
				   [WORD_FORM_OTHER] = 0 }) [n->stem_form];
	    v->penalty = ((uns[]) { [WORD_FORM_LEMMA] = morph_lemma_penalty,
				    [WORD_FORM_BASIC] = morph_basic_penalty,
				    [WORD_FORM_OTHER] = morph_other_penalty }) [n->stem_form];
	    if (n->unaccented)
	      {
		v->flags |= VF_PREACCENT;
		v->penalty += morph_preacc_penalty;
	      }
	    v->ignore_accents = (a->morph_mode >= MORPH_MODE_POSTACCENT && a->accent_mode == ACCENT_MODE_STRIP);
	  }
      }
#endif
}

static struct phrase *
phrase_build(struct query *q, struct lestrade_atom *a)
{
  uns cnt = 0;
  struct word *w;
  CLIST_WALK(w, a->words)
    if (w->pos != ~0U)
      cnt++;
  if (cnt < 2)
    return NULL;

  if (q->num_phrases >= max_phrases)
    query_error(q, LESTRADE_ERR_TOO_MANY_PHRASES, "Too many phrases");
  if (cnt > max_phrase_len)
    query_error(q, LESTRADE_ERR_PHRASE_TOO_LONG, "Phrase too long");

  struct phrase *p = mp_alloc_zero(q->pool, sizeof(*p));
  q->phrases[q->num_phrases++] = p;
  p->length = cnt;
  p->weight = a->weight;
  cnt = 0;
  uns last_pos = ~0U;
  CLIST_WALK(w, a->words)
    if (w->pos != ~0U)
      {
	p->words[cnt] = w;
	if (cnt)
	  p->relpos[cnt] = w->pos - last_pos;
	last_pos = w->pos;
	p->word_mask |= (1 << w->id);
	p->next_same_word[cnt] = p->word_to_idx[w->id];
	p->word_to_idx[w->id] = cnt+1;
	w->weight = 0;
	cnt++;
      }

  if (query_tracing(q, TRACE_WORDS))
    {
      DUMP("Added phrase %d: ", q->num_phrases-1);
      phrase_dump(p);
      DUMP("\n");
    }

  return p;
}

static void
words_expand_atom(struct query *q, struct lestrade_atom *a)
{
  lm_doc_start();
  word_current_query = q;
  word_current_atom = a;
  if (!lm_map_text(a->word))
    query_error(q, LESTRADE_ERR_WORD_TOO_LONG, "Word too long");
  words_dump_atom(q, a, "Mapped phrase");

  struct word *w;
  CLIST_WALK(w, a->words)
    if (w->word_class == WC_NORMAL)
	word_expand(q, a, w);
  words_dump_atom(q, a, "Expanded phrase");

  CLIST_WALK(w, a->words)
    if (w->pos != ~0U)
      {
	struct lestrade_expr *x = expr_new(q, EX_WORD);
	x->u.word = w;
	a->xlat = a->xlat ? expr_new_op(q, EX_AND, a->xlat, x) : x;
      }

  if (!a->xlat)
    a->xlat = expr_new(q, EX_IGNORE);
  else
    {
      struct phrase *p = phrase_build(q, a);
      if (p)
	{
	  struct lestrade_expr *x = expr_new(q, EX_PHRASE);
	  x->u.phrase = p;
	  a->xlat = expr_new_op(q, EX_AND, a->xlat, x);
	}
    }
}

static void
words_build_near(struct query *q, uns first_word_id, uns first_phrase_id)
{
  if (q->num_nears >= max_nears)
    return;
  if (q->num_phrases != first_phrase_id)
    return;

  uns cnt = 0;
  for (uns i=first_word_id; i<q->num_words; i++)
    if (q->words[i]->word_class == WC_NORMAL)
      cnt++;
  if (cnt < 2)
    return;

  struct phrase *n = mp_alloc_zero(q->pool, sizeof(*n));
  q->nears[q->num_nears++] = n;
  n->length = cnt;
  cnt = 0;
  for (uns i=first_word_id; i<q->num_words; i++)
    {
      struct word *w = q->words[i];
      if (w->word_class == WC_NORMAL)
	{
	  n->words[cnt] = w;
	  n->word_mask |= (1 << w->id);
	  n->next_same_word[cnt] = n->word_to_idx[w->id];
	  n->word_to_idx[w->id] = cnt+1;
	  cnt++;
	}
    }

  if (query_tracing(q, TRACE_WORDS))
    {
      DUMP("Added near %d: ", q->num_nears-1);
      phrase_dump(n);
      DUMP("\n");
    }
}

struct lestrade_expr *
words_expand(struct query *q, struct lestrade_expr *x)
{
  ASSERT(x->type == EX_MATCH);
  uns first_word_id = q->num_words;
  uns first_phrase_id = q->num_phrases;

  struct lestrade_atom *a;
  CLIST_WALK(a, x->u.match.atoms)
    words_expand_atom(q, a);
  words_build_near(q, first_word_id, first_phrase_id);

  return simple_fold(q, x);
}

void
words_global_init(void)
{
  lm_init();
#ifdef CONFIG_LANG
  lang_init_stemmers();
#endif
}
