/*
 * 	Lestrade Utilities - Index Dumper
 * 	
 * 	(c) 2005 Tomas Valla <tom@ucw.cz>
 */

#include "ucw/lib.h"
#include "ucw/getopt.h"
#include "lestrade/lestrade.h"
#include "custom/custom.h"
#include "ucw/fastbuf.h"
#include "lestrade/index.h"
#include "ucw/ff-unicode.h"
#include "ucw/ff-binary.h"
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <fcntl.h>

static struct fastbuf *stdoutfb;
#define OUT(x...) bprintf(stdoutfb, x)

static struct fastbuf *idxfb;

static struct index_header idx_hdr;

static void section(char *name)
{
	static int section_cnt;
	if (section_cnt++)
		OUT("\n");
	OUT("%s:\n", name);
}

static void dump_header(void)
{
	if (idx_hdr.magic == INDEX_MAGIC)
		section("Index header");
	else
		section("Temporary index header");
	OUT("Index format version:\t%02x%s\n", idx_hdr.version >> 24, ((idx_hdr.version >> 16) & 0xff) ? " large" : "");
	OUT("Custom index type:\t%02x version %02x\n", (idx_hdr.version >> 8) & 0xff, idx_hdr.version & 0xff);
	time_t t = idx_hdr.timestamp;
	OUT("Build time:\t\t%s", ctime(&t));
	OUT("Number of cards:\t%d\n", idx_hdr.num_cards);
	OUT("Custom data start:\t0x%x\n", idx_hdr.custom_start);
	OUT("Card attrs start:\t0x%x\n", idx_hdr.card_attr_start);
	OUT("References start:\t0x%x\n", idx_hdr.refs_start);
	OUT("Lexicon pointers start:\t0x%x\n", idx_hdr.lexicon_ptrs_start);
	OUT("Lexicon start:\t\t0x%x\n", idx_hdr.lexicon_start);
}

static void dump_zones(void)
{
	section("Lexicon zones");
	for (uns i=0; i<=MAX_WORD_BYTES+1; i++)
		OUT("Zone %2d:\t0x%x\n", i, idx_hdr.lexicon_zone[i]);
}

static void dump_custom_data(void)
{
	section("Custom data");
	bsetpos(idxfb, idx_hdr.custom_start);
	uns len = idx_hdr.card_attr_start - idx_hdr.custom_start;
	byte *buf = alloca(len);
	breadb(idxfb, buf, len);
	dump_custom_index_hdr(stdoutfb, (struct custom_index_hdr *) buf, len);
}

static void dump_card_attrs(void)
{
	struct card_attrs c;

	section("Card attributes");
	bsetpos(idxfb, idx_hdr.card_attr_start);
	for (uns i=0; i<idx_hdr.num_cards; i++) {
		breadb(idxfb, &c, sizeof(c));
		OUT("Card %d: ", i);
		dump_struct_card_attrs(stdoutfb, &c);
	}
}

static void dump_lexicon(void)
{
	uns i=0;
	uns l;
	byte wbuf[MAX_WORD_BYTES+1];
	index_pos_t start;

	section("Lexicon");
	bsetpos(idxfb, idx_hdr.lexicon_start);

	while ((l = bgetc(idxfb)) != ~0U) {
		i++;
		breadb(idxfb, wbuf, MIN(l, sizeof(wbuf)-1));
		wbuf[MIN(l, sizeof(wbuf)-1)] = 0;
		start = bgeto(idxfb);
		OUT("0x%x\t<%s> refchain 0x%x\n", i, wbuf, start);
	}
}

static void dump_refchains(void)
{
	u32 hdr, len;
	ucw_off_t fpos;
	uns c, pos, wt;

	section("Reference chains");
	bsetpos(idxfb, idx_hdr.refs_start);

	while ((fpos = btell(idxfb)) < (ucw_off_t) idx_hdr.lexicon_ptrs_start) {
		fpos -= idx_hdr.refs_start;
		OUT("Chain at position 0x%Lx\n", (long long)fpos);

		while (hdr = bgetl(idxfb)) {
			if (hdr & 15)
				len = hdr & 15;
			else
				len = bget_utf8(idxfb);
			OUT("\tOID 0x%x:", hdr >> 4);

			uns i = 0;
			while (i < len) {
				c = bgetc(idxfb);
				if (!(c & 0x80)) {
					wt = c >> 6;
					pos = c & 0x3f;
					i++;
				}
				else if(!(c & 0x40)) {
					wt = (c >> 3) & 7;
					pos = ((c & 7) << 8) | bgetc(idxfb);
					i += 2;
				}
				else {
					wt = (c >> 2) & 15;
					pos = ((c & 3) << 16) | (bgetc(idxfb) << 8) | bgetc(idxfb);
					i += 3;
				}
				OUT(" %d(%d)", pos, wt);
			}
			OUT("\n");
		}
	}
}

static void NONRET usage(byte *msg)
{
	if (msg)
		fputs(msg, stderr);
	fputs("\n\
Usage: lidxdump [<options>]\n\
-A		dump all\n\
-a		dump card attributes\n\
-c		dump custom data\n\
-f <index>	index filename\n\
-h		dump header\n\
-L		dump whole lexicon\n\
-R		dump all reference chains\n\
-z		dump lexicon zones\n\
", stderr);
	exit(1);
}
static char *options = "Aacf:hLRz";

static int header_dump;
static int custom_dump;
static int card_attrs_dump;
static int lexicon_dump;
static int refchains_dump;
static int zones_dump;

int main(int argc, char **argv)
{
	int opt;

	stdoutfb = bfdopen(1, 65536);

	log_init(argv[0]);
	while ((opt = cf_getopt(argc, argv, options, CF_NO_LONG_OPTS, NULL)) >= 0)
		switch (opt) {
		case 'A':
			header_dump = 1;
			custom_dump = 1;
			card_attrs_dump = 1;
			lexicon_dump = 1;
			refchains_dump = 1;
			zones_dump = 1;
			break;
		case 'a':
			card_attrs_dump = 1;
			break;
		case 'c':
			custom_dump = 1;
			break;
		case 'f':
			idxfb = bopen(optarg, O_RDONLY, 65536);
			break;
		case 'h':
			header_dump = 1;
			break;
		case 'L':
			lexicon_dump = 1;
			break;
		case 'R':
			refchains_dump = 1;
			break;
		case 'z':
			zones_dump = 1;
			break;
		default:
			usage("Bad options specified");
		}

	if (!idxfb)
		usage("No indexfile specified");

	/* Load header */
	breadb(idxfb, &idx_hdr, sizeof(idx_hdr));
	if (idx_hdr.magic != INDEX_MAGIC && idx_hdr.magic != INDEX_TEMP_MAGIC)
		die("Unrecognized index header (magic=%08x)", idx_hdr.magic);

	if (header_dump)
		dump_header();
	if (zones_dump)
		dump_zones();
	if (custom_dump)
		dump_custom_data();
	if (card_attrs_dump)
		dump_card_attrs();
	if (refchains_dump)
		dump_refchains();
	if (lexicon_dump)
		dump_lexicon();

	bclose(idxfb);
	bclose(stdoutfb);
	return 0;
}
